from .graphene_auth import broker, prototype_order, issue, reserve
from .config import NODES


def main():
    info = {
        "asset_id": "1.3.123",  # "1.3.x"
        "asset_name": "GATEWAY.XYZ",  # all caps
        "asset_precision": 5,  # int()
        "currency_id": "1.3.1234",  # "1.3.x"
        "currency_name": "GATEWAY.XYZA",  # all caps
        "currency_precision": 5,  # int()
        "issuer_id": "1.2.54321",  # "1.2.x"
        "issuer_public": "user",  # bitshares account name
        "issuer_private": "X" * 32,  # wif
    }

    issue(info, 10, "1.2.1234")
    reserve(info, 10)

    order = prototype_order(info)
    order["edicts"] = [{"op": "transfer", "amount": 10, "account_id": "1.2.789"}]
    broker(order)

    order = prototype_order(info)
    order["edicts"] = [{"op": "buy", "amount": 10, "price": 0.1, "expiration": 0}]
    broker(order)

    order = prototype_order(info)
    order["edicts"] = [{"op": "sell", "amount": 10, "price": 0.1, "expiration": 0}]
    broker(order)

    order = prototype_order(info)
    order["edicts"] = [{"op": "login"}]
    broker(order)


if __name__ == "__main__":
    main()



r"""
graphene_auth.py

  ____  _ _   ____  _                         
 | __ )(_) |_/ ___|| |__   __ _ _ __ ___  ___ 
 |  _ \| | __\___ \| '_ \ / _` | '__/ _ \/ __|
 | |_) | | |_ ___) | | | | (_| | | |  __/\__ \
 |____/|_|\__|____/|_| |_|\__,_|_|  \___||___/
       ____  _             _                  
      / ___|(_) __ _ _ __ (_)_ __   __ _      
      \___ \| |/ _` | '_ \| | '_ \ / _` |     
       ___) | | (_| | | | | | | | | (_| |     
      |____/|_|\__, |_| |_|_|_| |_|\__, |     
               |___/               |___/      


WTFPL litepresence.com Dec 2021 & squidKid-deluxe Jan 2024

BitShares ECDSA for Login, Buy, Sell, Cancel, Transfer, Issue, Reserve

"""
# DISABLE SELECT PYLINT TESTS
# pylint: disable=bad-continuation, broad-except, too-many-locals, too-many-statements
# pylint: disable=too-many-branches
#
# STANDARD PYTHON MODULES
import time  # hexidecimal to binary text
from decimal import Decimal as decimal
from multiprocessing import Process, Value  # convert back to PY variable

from .build_transaction import build_transaction
# GRAPHENE SIGNING MODULES
from .config import ATTEMPTS, JOIN, NODES, PROCESS_TIMEOUT
from .graphene_signing import (PrivateKey, serialize_transaction,
                               sign_transaction, verify_transaction)
from .rpc import rpc_broadcast_transaction, rpc_key_reference, wss_handshake
from .utilities import it, trace

# ISO8601 timeformat; 'graphene time'
ISO8601 = "%Y-%m-%dT%H:%M:%S%Z"


# Order creation helpers


def prototype_order(info, nodes=None):
    """
    Generates a prototype order, including a header with account and asset
    metadata. It requires a list of cached nodes and asset details to build
    the order structure. If no nodes are provided, it defaults to using the
    `NODES` listed in `config.py`.

    Parameters:
    info (dict): A dictionary containing order-specific information, notably:
                - `asset_id`: The ID of the asset being traded.
                - `asset_precision`: Precision for the asset.
                - `currency_id`: The ID of the associated currency (default is 0).
                - `currency_precision`: Precision for the currency (default is 0).
                - `account_id`: The ID of the issuer's account.
                - `account_name`: The issuer's public account name.
                - `wif`: The issuer's private key in Wallet Import Format (WIF).

    nodes (list, optional): A list of nodes to associate with the order. If 
                            not provided, defaults to `NODES`.

    Returns:
    dict: A dictionary representing a prototype order, containing the 
          `header` and `nodes`.  To make it not a prototype, simply add an "edicts" key
          with your order information.
    
    """
    if nodes is None:
        nodes = NODES
    header = {
        "asset_id": info["asset_id"],
        "asset_precision": info["asset_precision"],
        "currency_id": info.get("currency_id", 0),
        "currency_precision": info.get("currency_precision", 0),
        "account_id": info["account_id"],
        "account_name": info["account_name"],
        "wif": info["wif"],
    }
    order = {
        "header": header,
        "nodes": nodes,
    }
    return order


def issue(info, amount, account_id):
    """
    Put UIA.XYZ in user's BitShares wallet.
    """
    order = prototype_order(info)
    order["edicts"] = [{"op": "issue", "amount": amount, "account_id": account_id}]
    print(order["header"]["account_name"], order["header"]["asset_id"], order["edicts"])
    broker(order)


def reserve(info, amount):
    """
    Put UIA.XYZ into the reserve pool.
    """
    order = prototype_order(info)
    order["edicts"] = [{"op": "reserve", "amount": amount}]
    print(order["header"]["account_name"], order["header"]["asset_id"], order["edicts"])
    broker(order)


# Main process
def broker(order, broadcast=True):
    """
    Executes an authenticated operation (one of SUPPORTED_OPS) with robust error handling.

    This function acts as a timed, multiprocess wrapper for authorized operations. 
    It ensures that buy, sell, or cancel commands are executed promptly by enforcing a 
    timeout for each operation. If the operation does not complete within the 
    specified timeout, the process is terminated and restarted. This also helps 
    disconnect any hung websockets.

    Parameters:
    order (dict): A dictionary containing the details of the order to execute.
                  See prototype_order for more documentation.

    Returns:
    None

    Behavior:
    - The operation will attempt execution up to a defined number of times (`ATTEMPTS`).
    - Each attempt has a timeout duration of `PROCESS_TIMEOUT`.
    - If the operation still fails to execute within the timeout, it will be aborted.
    - After successful execution, the signal is set to 0 to indicate the process is complete.
    """
    signal = Value("i", 0)
    auth = Value("i", 0)
    iteration = 0
    while (iteration < ATTEMPTS) and not signal.value:
        iteration += 1
        print("\nmanualSIGNING authentication attempt:", iteration, time.ctime(), "\n")
        child = Process(target=execute, args=(signal, auth, order, broadcast))
        child.daemon = False
        child.start()
        if JOIN:  # means main script will not continue till child done
            child.join(PROCESS_TIMEOUT)
            child.terminate()

    return bool(auth.value)


def execute(signal, auth, order, broadcast):
    """
    #
    """

    def transact(rpc, order, auth):
        trx = build_transaction(rpc, order)
        # if there are any orders, perform ecdsa on serialized transaction
        if trx == -1:
            msg = it("red", "CURRENCY NOT PROVIDED")
        elif trx["operations"]:
            trx, message = serialize_transaction(rpc, trx)
            signed_tx = sign_transaction(trx, message, wif)
            if signed_tx is None:
                msg = it("red", "FAILED TO AUTHENTICATE ORDER")
                return msg
            signed_tx = verify_transaction(signed_tx, wif)
            # don't actaully broadcast login op, signing it is enough
            if order["edicts"][0]["op"] != "login" and broadcast:
                print(rpc_broadcast_transaction(rpc, signed_tx, order["header"]["client_order_id"]))
            auth.value = 1
            msg = it("green", ("EXECUTED ORDER" if broadcast else "SIGNED AND VERIFIED ORDER"))
        else:
            msg = it("red", "REJECTED ORDER")
        return msg
    rpc = wss_handshake()

    wif = order["header"]["wif"]
    start = time.time()
    # if this is just an authentication test, then there is no serialization / signing
    # just check that the private key references to the account id in question
    if order["edicts"][0]["op"] == "login":
        msg = it("red", "LOGIN FAILED")
        try:
            # instantitate a PrivateKey object
            private_key = PrivateKey(wif)
            # which contains an Address object
            address = private_key.address
            # which contains str(PREFIX) and a Base58(pubkey)
            # from these two, build a human terms "public key"
            public_key = address.prefix + str(address.pubkey)
            # get a key reference from that public key to 1.2.x account id
            key_reference_id = rpc_key_reference(rpc, public_key)[0][0]
            # extract the account id in the metanode
            account_id = order["header"]["account_id"]
            print("wif account id", key_reference_id)
            print("order account id", account_id)
            # if they match we're authenticated
            if account_id == key_reference_id:
                auth.value = 1
                msg = it("green", "AUTHENTICATED")
        except Exception:
            pass
    else:
        try:
            if (  # cancel all
                order["edicts"][0]["op"] == "cancel" and "1.7.X" in order["edicts"][0]["ids"]
            ):
                msg = it("red", "NO OPEN ORDERS")
                open_orders = True
                while open_orders:
                    metanode = peerplays_trustless_client()
                    open_orders = metanode["orders"]
                    ids = order["edicts"][0]["ids"] = [
                        order["orderNumber"] for order in open_orders
                    ]
                    if ids:
                        msg = transact(rpc, order, auth)
                    time.sleep(5)
            elif (  # cancel some
                order["edicts"][0]["op"] == "cancel" and "1.7.X" not in order["edicts"][0]["ids"]
            ):
                msg = it("red", "NO OPEN ORDERS")
                open_orders = True
                while open_orders:
                    metanode = peerplays_trustless_client()
                    open_orders = metanode["orders"]
                    ids = order["edicts"][0]["ids"] = [
                        order["orderNumber"]
                        for order in open_orders
                        if order in order["edicts"][0]["ids"]
                    ]
                    if ids:
                        msg = transact(rpc, order, auth)
                    time.sleep(5)

            else:  # all other order types
                msg = transact(rpc, order, auth)

        except Exception as error:
            trace(error)
    stars = it("yellow", "*" * (len(msg) + 17))
    msg = "manualSIGNING " + msg
    print("\n")
    print(stars + "\n    " + msg + "\n" + stars)
    print("\n")
    print("process elapsed: %.3f sec" % (time.time() - start), "\n\n")
    signal.value = 1



r"""
build_transaction.py

  ____  _ _   ____  _                         
 | __ )(_) |_/ ___|| |__   __ _ _ __ ___  ___ 
 |  _ \| | __\___ \| '_ \ / _` | '__/ _ \/ __|
 | |_) | | |_ ___) | | | | (_| | | |  __/\__ \
 |____/|_|\__|____/|_| |_|\__,_|_|  \___||___/
       ____  _             _                  
      / ___|(_) __ _ _ __ (_)_ __   __ _      
      \___ \| |/ _` | '_ \| | '_ \ / _` |     
       ___) | | (_| | | | | | | | | (_| |     
      |____/|_|\__, |_| |_|_|_| |_|\__, |     
               |___/               |___/      


WTFPL litepresence.com Dec 2021 & squidKid-deluxe Jan 2024

Build graphene transactions from human-readable edicts

"""

import itertools
# STANDARD PYTHON MODULES
import time
from binascii import unhexlify  # hexidecimal to binary text
from collections import OrderedDict
from decimal import Decimal as decimal
from struct import unpack_from  # convert back to PY variable

# GRAPHENE SIGNING MODULES
from .config import AUTOSCALE, CORE_FEES, DUST, KILL_OR_FILL, LIMIT
from .graphene_signing import ObjectId
from .rpc import (rpc_account_id, rpc_balances, rpc_block_number,
                  rpc_lookup_asset_symbols, rpc_open_orders, rpc_tx_fees)
from .utilities import fraction, it, to_iso_date

# MAX is 4294967295; year 2106 due to 4 byte unsigned integer
END_OF_TIME = 4 * 10**9  # about 75 years in future
# very little
SATOSHI = decimal("0.00000001")
# almost 1
SIXSIG = decimal("0.999999")


def graphenize_cancel(rpc, cancel_edicts, fees, order, account_name, tx_operations):
    """
    Translate cancel orders to graphene
    """
    for edict in cancel_edicts:
        if "ids" not in edict.keys():
            edict["ids"] = ["1.7.X"]
        if "1.7.X" in edict["ids"]:  # the "cancel all" signal
            # for cancel all op, we collect all open orders in 1 market
            edict["ids"] = rpc_open_orders(rpc, account_name, order["header"])
            print(it("yellow", str(edict)))
        for order_id in edict["ids"]:
            # confirm it is good 1.7.x format:
            order_id = str(order_id)
            aaa, bbb, ccc = order_id.split(".", 2)
            assert int(aaa) == 1
            assert int(bbb) == 7
            assert int(ccc) == float(ccc) > 0
            # create cancel fee ordered dictionary
            fee = OrderedDict([("amount", fees["cancel"]), ("asset_id", "1.3.0")])
            # create ordered operation dicitonary for this edict
            operation = [
                2,  # two means "Limit_order_cancel"
                OrderedDict(
                    [
                        ("fee", fee),
                        ("fee_paying_account", account_id),
                        ("order", order_id),
                        ("extensions", []),
                    ]
                ),
            ]
            # append the ordered dict to the trx operations list
            tx_operations.append(operation)
    return tx_operations


def graphenize_limit_orders(
    create_edicts, fees, asset, currency, account_id, tx_operations
):
    """
    Translate limit orders to graphene
    """
    asset_id, asset_precision = asset
    currency_id, currency_precision = currency
    for idx, edict in enumerate(create_edicts):
        price = edict["price"]
        amount = edict["amount"]
        op_exp = int(edict["expiration"])
        # convert zero expiration flag to "really far in future"
        if op_exp == 0:
            op_exp = END_OF_TIME
        op_expiration = to_iso_date(op_exp)
        # we'll use ordered dicts and put items in api specific order
        min_to_receive = OrderedDict({})
        amount_to_sell = OrderedDict({})
        # derive min_to_receive & amount_to_sell from price & amount
        # means SELLING currency RECEIVING assets
        if edict["op"] == "buy":
            min_to_receive["amount"] = int(amount * 10**asset_precision)
            min_to_receive["asset_id"] = asset_id
            amount_to_sell["amount"] = int(amount * price * 10**currency_precision)
            amount_to_sell["asset_id"] = currency_id
        # means SELLING assets RECEIVING currency
        if edict["op"] == "sell":
            min_to_receive["amount"] = int(amount * price * 10**currency_precision)
            min_to_receive["asset_id"] = currency_id
            amount_to_sell["amount"] = int(amount * 10**asset_precision)
            amount_to_sell["asset_id"] = asset_id
        # Limit_order_create fee ordered dictionary
        fee = OrderedDict([("amount", fees["create"]), ("asset_id", "1.3.0")])
        # create ordered dicitonary from each buy/sell operation
        operation = [
            1,
            OrderedDict(
                [
                    ("fee", fee),  # OrderedDict
                    ("seller", account_id),  # "a.b.c"
                    ("amount_to_sell", amount_to_sell),  # OrderedDict
                    ("min_to_receive", min_to_receive),  # OrderedDict
                    ("expiration", op_expiration),  # ISO8601
                    ("fill_or_kill", KILL_OR_FILL),  # bool
                    (
                        "extensions",
                        [],
                    ),  # always empty list for our purpose
                ]
            ),
        ]
        tx_operations.append(operation)
    return tx_operations


def graphenize_issue(
    issue_edicts, fees, asset_id, asset_precision, account_id, tx_operations
):
    """
    Translate issue orders to graphene
    """
    for idx, issue in enumerate(issue_edicts):
        # convert to graphene amount, asset_id type
        # class Asset_issue(GrapheneObject): # OPERATION ID 14 "asset_issue"
        fee = OrderedDict([("amount", fees["issue"]), ("asset_id", "1.3.0")])
        graphene_amount = issue["amount"] * 10**asset_precision
        amount_dict = OrderedDict([("amount", graphene_amount), ("asset_id", asset_id)])
        # issue ordered dictionary from each buy/sell operation
        operation = [
            14,
            OrderedDict(
                [
                    ("fee", fee),
                    ("issuer", account_id),
                    ("asset_to_issue", amount_dict),
                    ("issue_to_account", issue["account_id"]),
                    # ("memo", issue["memo"]), # FIXME memos are not yet implemented
                    ("extensions", []),
                ]
            ),
        ]
        tx_operations.append(operation)
    return tx_operations


def graphenize_reserve(
    reserve_edicts, fees, asset_id, asset_precision, account_id, tx_operations
):
    """
    Translate reserve orders to graphene
    """
    for idx, reserve in enumerate(reserve_edicts):
        # convert to graphene amount, asset_id type
        # class Asset_reserve(GrapheneObject): # OPERATION ID 15 "asset_reserve"
        fee = OrderedDict([("amount", fees["reserve"]), ("asset_id", "1.3.0")])
        graphene_amount = reserve["amount"] * 10**asset_precision
        amount_dict = OrderedDict([("amount", graphene_amount), ("asset_id", asset_id)])
        # reserve ordered dictionary from each buy/sell operation
        operation = [
            15,
            OrderedDict(
                [
                    ("fee", fee),
                    ("payer", account_id),
                    ("amount_to_reserve", amount_dict),
                    ("extensions", []),
                ]
            ),
        ]
        tx_operations.append(operation)
    return tx_operations


def graphenize_transfer(
    transfer_edicts, fees, asset_id, asset_precision, account_id, tx_operations
):
    """
    Translate transfer orders to graphene
    """
    for idx, transfer in enumerate(transfer_edicts):
        # convert to graphene amount, asset_id type
        # class Transfer(GrapheneObject): # OPERATION ID 0 "transfer"
        fee = OrderedDict([("amount", fees["transfer"]), ("asset_id", "1.3.0")])
        graphene_amount = int(transfer["amount"] * 10**asset_precision)
        amount_dict = OrderedDict([("amount", graphene_amount), ("asset_id", asset_id)])
        # transfer ordered dictionary from each buy/sell operation
        operation = [
            0,
            OrderedDict(
                [
                    ("fee", fee),
                    ("from", account_id),
                    ("to", transfer["account_id"]),
                    ("amount", amount_dict),
                    # ("memo", transfer["memo"]), # FIXME memos are not yet implemented
                    ("extensions", []),
                ]
            ),
        ]
        tx_operations.append(operation)
    return tx_operations


def graphenize_login(login_edicts, fees, account_id, tx_operations):
    """
    Translate login to dummy graphene cancel order
    """
    if login_edicts:
        fee = OrderedDict([("amount", 0), ("asset_id", "1.3.0")])
        operation = [
            2,
            OrderedDict(
                [
                    ("fee", fee),
                    ("fee_paying_account", account_id),
                    ("order", "1.7.0"),
                    ("extensions", []),
                ]
            ),
        ]
        tx_operations.append(operation)
    return tx_operations


def scale_limit_orders(
    rpc, order, asset_id, currency_id, account_name, buy_edicts, sell_edicts
):
    """
    Scale order size to funds on hand
    """
    if AUTOSCALE or CORE_FEES:
        currency, assets, bitshares = rpc_balances(rpc, account_name, order["header"])
        if AUTOSCALE and buy_edicts + sell_edicts:
            # autoscale buy edicts
            if buy_edicts:
                currency_value = 0
                # calculate total value of each amount in the order
                for idx, _ in enumerate(buy_edicts):
                    currency_value += (
                        buy_edicts[idx]["amount"] * buy_edicts[idx]["price"]
                    )
                # scale the order amounts to means
                scale = SIXSIG * currency / (currency_value + SATOSHI)
                if scale < 1:
                    print(
                        it("yellow", "ALERT: scaling buy edicts to means: %.3f" % scale)
                    )
                    for idx, _ in enumerate(buy_edicts):
                        buy_edicts[idx]["amount"] *= scale
            # autoscale sell edicts
            if sell_edicts:
                asset_total = 0
                # calculate total amount in the order
                for idx, _ in enumerate(sell_edicts):
                    asset_total += sell_edicts[idx]["amount"]
                scale = SIXSIG * assets / (asset_total + SATOSHI)
                # scale the order amounts to means
                if scale < 1:
                    print(
                        it(
                            "yellow",
                            "ALERT: scaling sell edicts to means: %.3f" % scale,
                        )
                    )
                    for idx, _ in enumerate(sell_edicts):
                        sell_edicts[idx]["amount"] *= scale
        save_core_fees(buy_edicts, sell_edicts, asset_id, currency_id)

    # after scaling recombine buy and sell
    create_edicts = buy_edicts + sell_edicts

    ###### Remove dust edicts
    if DUST and create_edicts:
        create_edicts2 = []
        dust = DUST * 100000 / 10**asset_precision
        for idx, edict in enumerate(create_edicts):
            if edict["amount"] > dust:
                create_edicts2.append(edict)
            else:
                print(
                    it("red", "WARN: removing dust threshold %s order" % dust),
                    edict,
                )
        create_edicts = create_edicts2[:]  # copy as new list
        del create_edicts2
    return create_edicts


def graphenize_swaps(swap_edicts, fees, asset, currency, account_id, tx_operations):
    asset_id, asset_precision = asset
    currency_id, currency_precision = currency
    for edict in swap_edicts:
        price = edict["price"]
        amount = edict["amount_to_sell"]
        op_exp = int(edict["expiration"])
        pool = edict["pool"]
        fee = OrderedDict([("amount", fees["swap"]), ("asset_id", "1.3.0")])
        # convert zero expiration flag to "really far in future"
        if op_exp == 0:
            op_exp = END_OF_TIME
        op_expiration = to_iso_date(op_exp)
        # we'll use ordered dicts and put items in api specific order
        min_to_receive = OrderedDict({})
        amount_to_sell = OrderedDict({})
        # derive min_to_receive & amount_to_sell from price & amount
        # means SELLING currency RECEIVING assets
        if edict["op"] == "swap":
            amount_to_sell["amount"] = int(amount * 10**currency_precision)
            amount_to_sell["asset_id"] = currency_id

            if price == 0:
                min_to_receive["amount"] = int(1)
                min_to_receive["asset_id"] = asset_id
            else:
                min_to_receive["amount"] = int(amount / price * 10**asset_precision)
                min_to_receive["asset_id"] = asset_id

        operation = [
            63,
            OrderedDict(
                [
                    ("fee", fee),  # Asset(kwargs["fee"])),
                    ("account", account_id),  # ObjectId(kwargs["account"], "account")),
                    ("pool", pool),  # ObjectId(kwargs["pool"], "liquidity_pool")),
                    (
                        "amount_to_sell",
                        amount_to_sell,
                    ),  # Asset(kwargs["amount_to_sell"])),
                    (
                        "min_to_receive",
                        min_to_receive,
                    ),  # Asset(kwargs["min_to_receive"])),
                    (
                        "extensions",
                        [],
                    ),  # Array([])),
                ]
            ),
        ]
        tx_operations.append(operation)
    return tx_operations


def graphenize_pool_creation(
    pool_create_edicts,
    fees,
    asset,
    currency,
    account_id,
    tx_operations,
):
    asset_id, asset_precision = asset
    currency_id, currency_precision = currency

    for pool_create_edict in pool_create_edicts:
        # op_exp = int(pool_create_edict["expiration"])
        operation = [
            59,
            OrderedDict(
                [
                    (
                        "fee",
                        # FIXME: ARRGH MAGIC NUMBERS IN MY CODE
                        OrderedDict([("amount", 5000000), ("asset_id", "1.3.0")]),
                    ),
                    ("account", account_id),
                    ("asset_a", asset_id),
                    ("asset_b", currency_id),
                    ("share_asset", pool_create_edict["share_asset"]),
                    (
                        "taker_fee_percent",
                        int(pool_create_edict["taker_fee_percent"] * 100),
                    ),
                    (
                        "withdrawal_fee_percent",
                        int(pool_create_edict["withdrawal_fee_percent"] * 100),
                    ),
                ]
            ),
        ]
        tx_operations.append(operation)
    return tx_operations


def graphenize_call(
    call_edicts,
    fees,
    asset,
    currency,
    account_id,
    tx_operations,
):
    asset_id, asset_precision = asset
    currency_id, currency_precision = currency
    # TRANSLATE CALL ORDERS TO GRAPHENE
    for call in call_edicts:
        # convert to graphene amount, asset_id type
        delta_debt = OrderedDict(
            [
                ("amount", int(call["debt_delta"] * 10**asset_precision)),
                ("asset_id", asset_id),
            ]
        )
        delta_collateral = OrderedDict(
            [
                ("amount", int(call["collateral_delta"] * 10**currency_precision)),
                ("asset_id", currency_id),
            ]
        )
        # call fee ordered dictionary of graphene amount, asset_id
        fee = OrderedDict([("amount", fees["call"]), ("asset_id", "1.3.0")])
        # call ordered dictionary from each buy/sell operation
        operation = [
            3,
            OrderedDict(
                [
                    ("fee", fee),
                    ("funding_account", account_id),
                    ("delta_collateral", delta_collateral),
                    ("delta_debt", delta_debt),
                    (
                        "extensions",
                        OrderedDict(
                            [("target_collateral_ratio", int(1000 * call["tcr"]))]
                        ),
                    ),
                ]
            ),
        ]
        # add this op to our list of ops
        tx_operations.append(operation)
    return tx_operations


def graphenize_pool_deposit(
    pool_deposit_edicts,
    fees,
    asset,
    currency,
    account_id,
    tx_operations,
):
    asset_id, asset_precision = asset
    currency_id, currency_precision = currency
    for pool_deposit in pool_deposit_edicts:
        # call fee ordered dictionary of graphene amount, asset_id
        fee = OrderedDict([("amount", fees["pool_deposit"]), ("asset_id", "1.3.0")])
        # call ordered dictionary from each buy/sell operation
        operation = [
            61,
            OrderedDict(
                [
                    ("fee", fee),
                    ("account", account_id),
                    ("pool", pool_deposit["pool"]),
                    (
                        "amount_a",
                        OrderedDict(
                            [
                                (
                                    "amount",
                                    int(
                                        pool_deposit["amount_a"] * 10**asset_precision
                                    ),
                                ),
                                ("asset_id", asset_id),
                            ]
                        ),
                    ),
                    (
                        "amount_b",
                        OrderedDict(
                            [
                                (
                                    "amount",
                                    int(
                                        pool_deposit["amount_b"]
                                        * 10**currency_precision
                                    ),
                                ),
                                ("asset_id", currency_id),
                            ]
                        ),
                    ),
                    ("extensions", []),
                ]
            ),
        ]
        # add this op to our list of ops
        tx_operations.append(operation)
    return tx_operations


def graphenize_fee_pool(
    fee_pool_edicts,
    fees,
    asset,
    currency,
    account_id,
    tx_operations,
):
    asset_id, asset_precision = asset
    currency_id, currency_precision = currency

    for fee_pool in fee_pool_edicts:
        # call fee ordered dictionary of graphene amount, asset_id
        fee = OrderedDict([("amount", 0), ("asset_id", "1.3.0")])
        # call ordered dictionary from each buy/sell operation
        operation = [
            47,
            OrderedDict(
                [
                    (
                        "fee",
                        OrderedDict(
                            [("amount", fees["fee_pool"]), ("asset_id", "1.3.0")]
                        ),
                    ),
                    ("issuer", fee_pool["issuer"]),
                    ("asset_id", fee_pool["asset_id"]),
                    (
                        "amount_to_claim",
                        OrderedDict(
                            [
                                ("amount", fee_pool["amount_to_claim"]),
                                ("asset_id", "1.3.0"),
                            ]
                        ),
                    ),  #
                    # ("amount_to_claim", OrderedDict([("amount", 217), ("asset_id", "1.3.0")])),
                    ("extensions", []),
                ]
            ),
        ]
        # add this op to our list of ops
        tx_operations.append(operation)
    return tx_operations


def save_core_fees(buy_edicts, sell_edicts, asset_id, currency_id):
    """
    Always save last two bitshares for fees
    """
    if CORE_FEES and (
        buy_edicts + sell_edicts and ("1.3.0" in [asset_id, currency_id])
    ):
        # print(bitshares, 'BTS balance')
        # when BTS is the currency don't spend the last 2
        if currency_id == "1.3.0" and buy_edicts:
            bts_value = 0
            # calculate total bts value of each amount in the order
            for idx, _ in enumerate(buy_edicts):
                bts_value += buy_edicts[idx]["amount"] * buy_edicts[idx]["price"]
            # scale the order amounts to save last two bitshares
            scale = SIXSIG * max(0, (bitshares - 2)) / (bts_value + SATOSHI)
            if scale < 1:
                print(it("yellow", "ALERT: scaling buy edicts for fees: %.4f" % scale))
                for idx, _ in enumerate(buy_edicts):
                    buy_edicts[idx]["amount"] *= scale
        # when BTS is the asset don't sell the last 2
        if asset_id == "1.3.0" and sell_edicts:
            bts_total = 0
            # calculate total of each bts amount in the order
            for idx, _ in enumerate(sell_edicts):
                bts_total += sell_edicts[idx]["amount"]
            scale = SIXSIG * max(0, (bitshares - 2)) / (bts_total + SATOSHI)
            # scale the order amounts to save last two bitshares
            if scale < 1:
                print(
                    it(
                        "yellow",
                        "ALERT: scaling sell edicts for fees: %.4f" % scale,
                    )
                )
                for idx, _ in enumerate(sell_edicts):
                    sell_edicts[idx]["amount"] *= scale
    return buy_edicts, sell_edicts


def graphenize_asset_create(
    asset_create_edicts,
    fees,
    asset,
    currency,
    account_id,
    tx_operations,
):
    asset_id, asset_precision = asset
    currency_id, currency_precision = currency
    for asset_create_edict in asset_create_edicts:
        common_options = asset_create_edict["common_options"]
        operation = [
            10,
            OrderedDict(
                [
                    # FIXME: MORE magic numbers? really?
                    #                               96543404
                    # 96555555
                    # 96605000
                    ("fee", OrderedDict([("amount", 96700000), ("asset_id", "1.3.0")])),
                    ("issuer", asset_create_edict["issuer"]),
                    ("symbol", asset_create_edict["symbol"]),
                    ("precision", asset_create_edict["precision"]),
                    (
                        "common_options",
                        OrderedDict(
                            [
                                ("max_supply", common_options["max_supply"]),
                                (
                                    "market_fee_percent",
                                    common_options["market_fee_percent"],
                                ),
                                (
                                    "max_market_fee",
                                    common_options["max_market_fee"],
                                ),
                                (
                                    "issuer_permissions",
                                    common_options["issuer_permissions"],
                                ),
                                ("flags", common_options["flags"]),
                                (
                                    "core_exchange_rate",
                                    common_options["core_exchange_rate"],
                                ),
                                (
                                    "whitelist_authorities",
                                    common_options["whitelist_authorities"],
                                ),
                                (
                                    "blacklist_authorities",
                                    common_options["blacklist_authorities"],
                                ),
                                (
                                    "whitelist_markets",
                                    common_options["whitelist_markets"],
                                ),
                                (
                                    "blacklist_markets",
                                    common_options["blacklist_markets"],
                                ),
                                ("description", common_options["description"]),
                                ("extensions", []),
                            ]
                        ),
                    ),
                    (
                        "is_prediction_market",
                        asset_create_edict["is_prediction_market"],
                    ),
                    ("extensions", []),
                ]
            ),
        ]
        tx_operations.append(operation)
    return tx_operations


def graphenize_publish(publish_edicts, fees, rpc, account_id, tx_operations):
    """
    TRANSLATE PRICE FEED PUBLISH ORDERS TO GRAPHENE
    """
    symbols = list(
        set(
            itertools.chain(
                *[[i["currency_name"], i["asset_name"]] for i in publish_edicts]
            )
        )
    )
    symbol_data = dict(zip(symbols, rpc_lookup_asset_symbols(rpc, symbols)))
    for edict in publish_edicts:
        # make external rpc for id and precision for the asset
        asset_id, asset_precision = (
            symbol_data[edict["asset_name"]]["id"],
            symbol_data[edict["asset_name"]]["precision"],
        )
        # repeat for currency
        currency_id, currency_precision = (
            symbol_data[edict["currency_name"]]["id"],
            symbol_data[edict["currency_name"]]["precision"],
        )
        # adjust settlment price to graphene asset and currency precisions
        adj_settlement = (
            edict["settlement_price"] * 10**asset_precision / 10**currency_precision
        )
        if edict["currency_name"] == "BTS":
            adj_core = adj_settlement
        else:
            adj_core = edict["core_price"] * 10**asset_precision / 10**5
        # FEE ORDERED DICT
        # create publication fee ordered dict
        fee = OrderedDict([("amount", fees["publish"]), ("asset_id", "1.3.0")])
        # SETTLEMENT ORDERED DICT
        # convert settlement price to a base/quote fraction
        s_base = fraction(adj_settlement)["base"]
        s_quote = fraction(adj_settlement)["quote"]
        # create a settlement-base price ordered dict
        settlement_base = OrderedDict([("amount", s_base), ("asset_id", asset_id)])
        # create a quote price ordered dict used w/ settlement base
        settlement_quote = OrderedDict([("amount", s_quote), ("asset_id", currency_id)])
        # combine settlement base and quote price
        settlement_price = OrderedDict(
            [("base", settlement_base), ("quote", settlement_quote)]
        )
        # CORE ORDERED DICT
        # convert core price to a base/quote fraction and multiply base by CER coeff
        c_base = int(fraction(adj_core)["base"] * edict["CER"])
        c_quote = fraction(adj_core)["quote"]
        # create a core-base price ordered dict
        core_base = OrderedDict(
            [
                ("amount", c_base),
                ("asset_id", asset_id),
            ]
        )
        # create a quote price ordered dict used w/ core base
        core_quote = OrderedDict([("amount", c_quote), ("asset_id", "1.3.0")])
        # combine core base and quote price
        core_exchange_rate = OrderedDict([("base", core_base), ("quote", core_quote)])
        # /bitshares/bitshares.py
        feed = OrderedDict(
            (
                [
                    # https://www.finra.org/rules-guidance/key-topics/margin-accounts
                    ("settlement_price", settlement_price),
                    (
                        "maintenance_collateral_ratio",
                        edict["MCR"],  # use graphene precision
                    ),
                    ("maximum_short_squeeze_ratio", edict["MSSR"]),
                    ("core_exchange_rate", core_exchange_rate),
                ]
            )
        )
        operation = [
            19,  # nineteen means "Publish Price Feed"
            OrderedDict(
                [
                    ("fee", fee),
                    ("publisher", account_id),
                    ("asset_id", asset_id),
                    ("feed", feed),
                    ("extensions", []),
                ]
            ),
        ]
        tx_operations.append(operation)
    return tx_operations


def graphenize_add_producer(producer_edicts, fees, account_id, tx_operations):
    """
    TRANSLATE ADD PRODUCER ORDER TO GRAPHENE
    """
    for edict in producer_edicts:
        fee = OrderedDict([("amount", fees["add_producer"]), ("asset_id", "1.3.0")])
        operation = [
            13,  # thirteen means "edit the price feed producer list"
            OrderedDict(
                [
                    ("fee", fee),
                    ("issuer", account_id),
                    ("asset_to_update", edict["asset_id"]),
                    ("new_feed_producers", edict["producer_ids"]),
                    ("extensions", []),
                ]
            ),
        ]
        tx_operations.append(operation)
    return tx_operations


def build_transaction(rpc, order):
    """
    # this performs incoming limit order api conversion
    # from human terms to graphene terms

    # humans speak:
     - account name, asset name, order number
     - decimal amounts, rounded is just fine
     - buy/sell/cancel
     - amount of assets
     - price in currency

    # graphene speaks:
     - 1.2.x, 1.3.x, 1.7.x
     - only in integers
     - create/cancel
     - min_to_receive/10^receiving_precision
     - amount_to_sell/10^selling_precision

    # build_transaction speaks:
     - list of buy/sell/cancel human terms edicts any order in
     - validated data request
     - autoscale amounts if out of budget
     - autoscale amounts if spending last bitshare
     - bundled cancel/buy/sell transactions out; cancel first
     - prevent inadvertent huge number of orders
     - do not place orders for dust amounts
    """
    # VALIDATE INCOMING DATA
    for key, expected_type in [("edicts", list), ("nodes", list), ("header", dict)]:
        if not isinstance(order[key], expected_type):
            raise ValueError(
                f"order parameter '{key}' must be of type {expected_type.__name__}:"
                f" {order[key]}; is {type(order[key]).__name__}"
            )

    # the location of the decimal place must be provided by order
    asset_precision = int(order["header"].get("asset_precision", 5))
    asset_id = str(order["header"].get("asset_id", "1.3.0"))

    account_name = str(order["header"]["account_name"])
    account_id = str(
        order["header"].get("account_id", rpc_account_id(rpc, account_name))
    )

    currency_precision = int(order["header"].get("currency_precision", 0))
    currency_id = str(order["header"].get("currency_id", 0))

    checks = [account_id, asset_id]

    # perform checks on currency for limit orders
    if order["edicts"][0]["op"] in ["buy", "sell"]:
        checks.append(currency_id)
        if not currency_precision or not currency_id:
            return -1
    # validate a.b.c identifiers of account id and asset ids
    for check in checks:
        ObjectId(check)
    # GATHER TRANSACTION HEADER DATA
    # fetch block data via websocket request
    block = rpc_block_number(rpc)
    ref_block_num = block["head_block_number"] & 0xFFFF
    ref_block_prefix = unpack_from("<I", unhexlify(block["head_block_id"]), 4)[0]
    # fetch limit order create and cancel fee via websocket request
    fees = rpc_tx_fees(rpc, account_id)
    # establish transaction expiration
    tx_expiration = to_iso_date(int(time.time() + 120))
    # initialize tx_operations list
    tx_operations = []

    # Sort incoming edicts by type
    edict_types = {
        "buy": [],
        "sell": [],
        "cancel": [],
        "swap": [],
        "create_pool": [],
        "create_asset": [],
        "transfer": [],
        "reserve": [],
        "issue": [],
        "login": [],
        "call": [],
        "pool_deposit": [],
        "fee_pool": [],
        "publish": [],
        "add_producer": [],
        # TODO
        # add those from `shorting attack avenger`
    }

    for edict in order["edicts"]:
        if edict["op"] in edict_types:
            print(it("yellow", str({k: str(v) for k, v in edict.items()})))
            edict_types[edict["op"]].append(edict)

    # convert to decimal type

    for op_type in ("buy", "sell"):
        for idx, item in enumerate(edict_types[op_type]):
            edict_types[op_type][idx]["amount"] = decimal(item["amount"])
            edict_types[op_type][idx]["price"] = decimal(item["price"])

    tx_operations = graphenize_cancel(
        rpc, edict_types["cancel"], order, account_name, fees, tx_operations
    )
    if any([edict_types["sell"], edict_types["buy"]]):
        scaled_limit_orders = scale_limit_orders(
            rpc,
            order,
            asset_id,
            currency_id,
            account_name,
            edict_types["buy"],
            edict_types["sell"],
        )
        tx_operations = graphenize_limit_orders(
            scaled_limit_orders,
            fees,
            [asset_id, asset_precision],
            [currency_id, currency_precision],
            account_id,
            tx_operations,
        )

    # fmt: off
    # Define a list of operations with their corresponding function and parameters
    operations = [
        ("swap", graphenize_swaps, [[asset_id, asset_precision], [currency_id, currency_precision]]),
        ("create_pool", graphenize_pool_creation, [[asset_id, asset_precision], [currency_id, currency_precision]]),
        ("create_asset", graphenize_asset_create, [[asset_id, asset_precision], [currency_id, currency_precision]]),
        ("call", graphenize_call, [[asset_id, asset_precision], [currency_id, currency_precision]]),
        ("pool_deposit", graphenize_pool_deposit, [[asset_id, asset_precision], [currency_id, currency_precision]]),
        ("fee_pool", graphenize_fee_pool, [[asset_id, asset_precision], [currency_id, currency_precision]]),
        ("transfer", graphenize_transfer, [asset_id, asset_precision]),
        ("issue", graphenize_issue, [asset_id, asset_precision]),
        ("reserve", graphenize_reserve, [asset_id, asset_precision]),
        ("publish", graphenize_publish, [rpc]),
        ("add_producer", graphenize_add_producer, []),
        ("login", graphenize_login, []),
    ]
    # fmt: on

    # Iterate over the operations and apply them to tx_operations
    for name, func, params in operations:
        tx_operations = func(
            edict_types[name], fees, *params, account_id, tx_operations
        )

    # prevent inadvertent huge number of orders
    tx_operations = tx_operations[:LIMIT]
    # the trx is just a regular dictionary we will convert to json later
    # the operations themselves must still be an OrderedDict
    trx = {
        "ref_block_num": ref_block_num,
        "ref_block_prefix": ref_block_prefix,
        "expiration": tx_expiration,
        "operations": tx_operations,
        "signatures": [],
        "extensions": [],
    }
    return trx



r"""
rpc.py

  ____  _ _   ____  _                         
 | __ )(_) |_/ ___|| |__   __ _ _ __ ___  ___ 
 |  _ \| | __\___ \| '_ \ / _` | '__/ _ \/ __|
 | |_) | | |_ ___) | | | | (_| | | |  __/\__ \
 |____/|_|\__|____/|_| |_|\__,_|_|  \___||___/
       ____  _             _                  
      / ___|(_) __ _ _ __ (_)_ __   __ _      
      \___ \| |/ _` | '_ \| | '_ \ / _` |     
       ___) | | (_| | | | | | | | | (_| |     
      |____/|_|\__, |_| |_|_|_| |_|\__, |     
               |___/               |___/      


WTFPL litepresence.com Dec 2021 & squidKid-deluxe Jan 2024

Collection of functions that interact with BitShares nodes using a WebSocket connection.

"""
# DISABLE SELECT PYLINT TESTS
# pylint: disable=broad-except

# STANDARD PYTHON MODULES
import json
import time
from decimal import Decimal as decimal
from random import shuffle

# THIRD PARTY MODULES
from websocket import create_connection as wss  # handshake to node

# GRAPHENE SIGNING MODULES
from .config import HANDSHAKE_TIMEOUT, NODES
from .utilities import trace


def wss_handshake(rpc=None):
    """
    create a wss handshake in less than X seconds, else try again
    """
    shuffle(NODES)
    handshake = HANDSHAKE_TIMEOUT + 1
    while handshake > HANDSHAKE_TIMEOUT:
        try:
            try:
                if rpc is not None:
                    rpc.close()  # attempt to close open stale connection
            except Exception:
                pass
            start = time.time()
            NODES.append(NODES.pop(0))  # rotate list
            node = NODES[0]
            rpc = wss(node, timeout=HANDSHAKE_TIMEOUT)
            handshake = time.time() - start
        except Exception:
            continue
    return rpc


def wss_query(rpc, params, client_order_id=1):
    """
    this definition will place all remote procedure calls (RPC)
    """

    for _ in range(10):
        try:
            # print(it('purple','RPC ' + params[0])('cyan',params[1]))
            # this is the 4 part format of EVERY rpc request
            # params format is ["location", "object", []]
            query = json.dumps(
                {"method": "call", "params": params, "jsonrpc": "2.0", "id": client_order_id}
            )
            # print(query)
            # rpc is the rpc connection created by wss_handshake()
            # we will use this connection to send query and receive json
            rpc.send(query)
            ret = json.loads(rpc.recv())
            try:
                ret = ret["result"]  # if there is result key take it
            except Exception:
                pass
            # print(ret)
            # print('elapsed %.3f sec' % (time.time() - start))
            return ret
        except Exception as error:
            try:  # attempt to terminate the connection
                rpc.close()
            except Exception:
                pass
            trace(error)  # tell me what happened
            # switch NODES
            rpc = wss_handshake(rpc)
            continue


def rpc_block_number(rpc):
    """
    block number and block prefix
    """
    return wss_query(rpc, ["database", "get_dynamic_global_properties", []])


def rpc_account_id(rpc, account_name):
    """
    given an account name return an account id
    """
    ret = wss_query(rpc, ["database", "lookup_accounts", [account_name, 1]])
    return ret[0][1]


def rpc_get_account(rpc, account_name):
    """
    given an account name return an account id
    """
    ret = wss_query(rpc, ["database", "get_account_by_name", [account_name, 1]])
    return ret


def rpc_tx_fees(rpc, account_id):
    # returns fee for limit order create and cancel without 10^precision

    query = [
        "database",
        "get_required_fees",
        [
            [
                ["0", {"from": str(account_id)}],
                ["1", {"from": str(account_id)}],
                ["2", {"from": str(account_id)}],
                ["3", {"from": str(account_id)}],
                ["10", {"from": str(account_id)}],
                ["13", {"from": str(account_id)}],  # add a feed producer
                ["14", {"from": str(account_id)}],
                ["19", {"from": str(account_id)}],  # publish a price feed
                ["15", {"from": str(account_id)}],
                ["47", {"from": str(account_id)}],
                ["59", {"from": str(account_id)}],
                ["61", {"from": str(account_id)}],
                ["63", {"from": str(account_id)}],
            ],
            "1.3.0",
        ],
    ]
    ret = wss_query(rpc, query)
    final_ret = {
        "transfer": ret[0]["amount"],
        "create": ret[1]["amount"],
        "cancel": ret[2]["amount"],
        "call": ret[3]["amount"],
        "asset_create": ret[4]["amount"],
        "add_producer": ret[5]["amount"],
        "issue": ret[6]["amount"],
        "publish": ret[7]["amount"],
        "asset_reserve": ret[8]["amount"],
        "fee_pool": ret[9]["amount"],
        "pool_create": ret[10]["amount"],
        "pool_deposit": ret[11]["amount"],
        "swap": ret[12]["amount"],
    }
    return final_ret


def rpc_balances(rpc, account_name, pair):
    """
    account balances
    """
    balances = wss_query(rpc, 
        [
            "database",
            "get_named_account_balances",
            [account_name, [pair["currency_id"], pair["asset_id"], "1.3.0"]],
        ]
    )

    for balance in balances:
        if balance["asset_id"] == pair["currency_id"]:
            currency = decimal(balance["amount"]) / 10 ** pair["currency_precision"]
        if balance["asset_id"] == pair["asset_id"]:
            assets = decimal(balance["amount"]) / 10 ** pair["asset_precision"]
        if balance["asset_id"] == "1.3.0":
            bitshares = decimal(balance["amount"]) / 10**5

    # print(currency, assets, bitshares)
    return currency, assets, bitshares


def rpc_open_orders(rpc, account_name, pair):
    """
    return a list of open orders, for one account, in one market
    """
    ret = wss_query(rpc, ["database", "get_full_accounts", [[account_name], "false"]])
    try:
        limit_orders = ret[0][1]["limit_orders"]
    except Exception:
        limit_orders = []
    market = [pair["currency_id"], pair["asset_id"]]
    orders = []
    for order in limit_orders:
        base_id = order["sell_price"]["base"]["asset_id"]
        quote_id = order["sell_price"]["quote"]["asset_id"]
        if (base_id in market) and (quote_id in market):
            orders.append(order["id"])
    return orders


def rpc_key_reference(rpc, public_key):
    """
    given public key return account id
    """
    return wss_query(rpc, ["database", "get_key_references", [[public_key]]])


def rpc_get_transaction_hex_without_sig(rpc, trx):
    """
    use this to verify the manually serialized trx buffer
    """
    ret = wss_query(rpc, ["database", "get_transaction_hex_without_sig", [trx]])
    return bytes(ret, "utf-8")


def rpc_get_transaction_hex(rpc, trx):
    """
    use this to verify the manually serialized trx buffer
    """
    ret = wss_query(rpc, ["database", "get_transaction_hex", [trx]])
    return bytes(ret, "utf-8")


def rpc_get_objects(rpc, obj_id):
    """
    use this to verify the manually serialized trx buffer
    """
    return wss_query(rpc, ["database", "get_objects", [[obj_id]]])[0]


def rpc_broadcast_transaction(rpc, trx, client_order_id=1):
    """
    upload the signed transaction to the blockchain
    """
    ret = wss_query(rpc, ["network_broadcast", "broadcast_transaction", [trx]], client_order_id)

    print(json.dumps(ret, indent=4))

    return ret


def rpc_lookup_asset_symbols(rpc, asset):  # DONE
    """
    Given asset names return asset ids and precisions
    """
    if isinstance(asset, str):
        asset = [asset]

    return wss_query(rpc, ["database", "lookup_asset_symbols", [asset]])


def unit_test():
    """
    test functionality of select definitions
    """
    rpc = wss_handshake()
    rpc_get_account(rpc, "litepresence1")


if __name__ == "__main__":
    unit_test()



r"""
utilities.py

  ____  _ _   ____  _                         
 | __ )(_) |_/ ___|| |__   __ _ _ __ ___  ___ 
 |  _ \| | __\___ \| '_ \ / _` | '__/ _ \/ __|
 | |_) | | |_ ___) | | | | (_| | | |  __/\__ \
 |____/|_|\__|____/|_| |_|\__,_|_|  \___||___/
       ____  _             _                  
      / ___|(_) __ _ _ __ (_)_ __   __ _      
      \___ \| |/ _` | '_ \| | '_ \ / _` |     
       ___) | | (_| | | | | | | | | (_| |     
      |____/|_|\__, |_| |_|_|_| |_|\__, |     
               |___/               |___/      


WTFPL litepresence.com Dec 2021 & squidKid-deluxe Jan 2024

UTILITIES

"""

# DISABLE SELECT PYLINT TESTS
# pylint: disable=invalid-name

import os
import sys
from calendar import timegm
from datetime import datetime
from time import strptime
from traceback import format_exc

# ISO8601 timeformat; 'graphene time'
ISO8601 = "%Y-%m-%dT%H:%M:%S%Z"


def disable_print():
    """
    temporarily disable printing
    """
    sys.stdout = open(os.devnull, "w")


def enable_print():
    """
    re-enable printing
    """
    sys.stdout = sys.__stdout__


def trace(error):
    """
    print stack trace upon exception
    """
    msg = str(type(error).__name__) + "\n"
    msg += str(error.args) + "\n"
    msg += str(format_exc()) + "\n"
    print(msg)


def it(style, text):
    """
    Text coloring
    """
    emphasis = {
        "red": 91,
        "green": 92,
        "yellow": 93,
        "blue": 94,
        "purple": 95,
        "cyan": 96,
    }

    return ("\033[%sm" % emphasis[style]) + str(text) + "\033[0m"


def to_iso_date(unix):
    """
    returns iso8601 datetime given unix epoch
    """
    return datetime.utcfromtimestamp(int(unix)).isoformat()


def from_iso_date(iso):
    """
    returns unix epoch given iso8601 datetime
    """
    return int(timegm(strptime((iso + "UTC"), ISO8601)))



def fraction(num):
    """
    convert float to limited precision least common denominator fraction
    """
    iteration = 0
    den = 1
    while True:  # move decimal place by factor of 10
        iteration += 1
        num *= 10
        den *= 10
        # escape when numerator is integer or denomenator approaches double long int
        if (int(num) == num) or (den == 10**14):
            break
    # ensure numerator is now an integer
    num = int(num)
    while True:  # remove common factors of 2
        iteration += 1
        if int(num / 2) != num / 2 or int(den / 2) != den / 2:
            break
        num /= 2
        den /= 2
    while True:  # remove common factors of 5
        iteration += 1
        if int(num / 5) != num / 5 or int(den / 5) != den / 5:
            break
        num /= 5
        den /= 5
    return {"base": int(num), "quote": int(den), "iteration": iteration}



# -*- coding: utf-8 -*-
import hashlib
import struct
import sys

from binascii import hexlify, unhexlify


try:
    from Cryptodome.Cipher import AES
except ImportError:
    try:
        from Crypto.Cipher import AES
    except ImportError:
        raise ImportError("Missing dependency: pyCryptodome")


def get_shared_secret(priv, pub):
    """Derive the share secret between ``priv`` and ``pub``

    :param `Base58` priv: Private Key
    :param `Base58` pub: Public Key
    :return: Shared secret
    :rtype: hex

    The shared secret is generated such that::

        Pub(Alice) * Priv(Bob) = Pub(Bob) * Priv(Alice)

    """
    pub_point = pub.point()
    priv_point = int(repr(priv), 16)
    res = pub_point * priv_point
    res_hex = "%032x" % res.x()
    # Zero padding
    res_hex = "0" * (64 - len(res_hex)) + res_hex
    return res_hex


def init_aes(shared_secret, nonce):
    """Initialize AES instance

    :param hex shared_secret: Shared Secret to use as encryption key
    :param int nonce: Random nonce
    :return: AES instance
    :rtype: AES

    """
    " Shared Secret "
    ss = hashlib.sha512(unhexlify(shared_secret)).digest()
    " Seed "
    seed = bytes(str(nonce), "ascii") + hexlify(ss)
    seed_digest = hexlify(hashlib.sha512(seed).digest()).decode("ascii")
    " AES "
    key = unhexlify(seed_digest[0:64])
    iv = unhexlify(seed_digest[64:96])
    return AES.new(key, AES.MODE_CBC, iv)


def _pad(s, BS):
    numBytes = BS - len(s) % BS
    return s + numBytes * struct.pack("B", numBytes)


def _unpad(s, BS):
    count = s[-1]
    if s[-count::] == count * struct.pack("B", count):
        return s[:-count]
    return s


def encode_memo(priv, pub, nonce, message):
    """Encode a message with a shared secret between Alice and Bob

    :param PrivateKey priv: Private Key (of Alice)
    :param PublicKey pub: Public Key (of Bob)
    :param int nonce: Random nonce
    :param str message: Memo message
    :return: Encrypted message
    :rtype: hex

    """
    shared_secret = get_shared_secret(priv, pub)
    aes = init_aes(shared_secret, nonce)
    " Checksum "
    raw = bytes(message, "utf8")
    checksum = hashlib.sha256(raw).digest()
    raw = checksum[0:4] + raw
    " Padding "
    raw = _pad(raw, 16)
    " Encryption "
    return hexlify(aes.encrypt(raw)).decode("ascii")


def decode_memo(priv, pub, nonce, message):
    """Decode a message with a shared secret between Alice and Bob

    :param PrivateKey priv: Private Key (of Bob)
    :param PublicKey pub: Public Key (of Alice)
    :param int nonce: Nonce used for Encryption
    :param bytes message: Encrypted Memo message
    :return: Decrypted message
    :rtype: str
    :raise ValueError: if message cannot be decoded as valid UTF-8
           string

    """
    shared_secret = get_shared_secret(priv, pub)
    aes = init_aes(shared_secret, nonce)
    " Encryption "
    raw = bytes(message, "ascii")
    cleartext = aes.decrypt(unhexlify(raw))
    " Checksum "
    checksum = cleartext[0:4]
    message = cleartext[4:]
    message = _unpad(message, 16)
    " Verify checksum "
    check = hashlib.sha256(message).digest()[0:4]
    if check != checksum:  # pragma: no cover
        raise ValueError("checksum verification failure")
    return message.decode("utf8")
